import fnmatch
import datetime
from typing import Any, Dict

import pytz
from io import BytesIO, StringIO
import pandas as pd

from components.Utils import get_type_file
from enums.FileTypeEnum import FileTypeEnum

from airflow.providers.amazon.aws.hooks.s3 import S3Hook
import logging

logger = logging.getLogger()


def get_df_from_s3(conn: str, bucket: str, key: str, delimiter: str, base_date: datetime.date,
                   interval: str) -> Dict[str, Any]:
    response = {'filename': "", 'df': None}
    dataframe = None
    try:
        s3_data = get_data_from_s3(conn, bucket, key, base_date, interval)
        logger.info(f"ARCHIVO EXTRAIDO: {s3_data}")
        if s3_data["filename"] == "":
            raise Exception(f"No se encontró archivo para el key: {key} y fecha base {base_date} en intervalo {interval}")
        response.update({'filename': s3_data["filename"]})
        file_type = get_type_file(s3_data["filename"])
        if file_type == FileTypeEnum.EXCEL:
            dataframe = pd.read_excel(s3_data["data"], engine="openpyxl", dtype='object')
        elif file_type == FileTypeEnum.OLD_EXCEL:
            dataframe = pd.read_excel(s3_data["data"], engine="xlrd", dtype='object')
        elif file_type == FileTypeEnum.TEXT or file_type == FileTypeEnum.CSV:
            str_data = str(s3_data["data"].getvalue(), encoding='UTF-8', errors='ignore')
            data = StringIO(str_data)
            dataframe = pd.read_csv(data, sep=delimiter, dtype='object')
        response.update({'df': dataframe})
    except Exception as e:
        logger.error(f"Error trayendo y transformando a DataFrame desde S3. {e}")
    return response


def get_data_from_s3(conn: str, bucket: str, key: str, base_date: datetime.date, interval: str) -> Dict[str, Any]:
    result = {'filename': '', 'data': BytesIO()}
    utc = pytz.UTC
    try:
        if key.rfind("/") != -1:
            prefix = key[:key.rfind("/")+1]
        else:
            prefix = ""
        s3_hook = S3Hook(conn)
        files = s3_hook.list_keys(bucket, prefix)
        last_key = ("", base_date)
        # Colocar una fecha muy atrás como base
        for file_key in files:
            if fnmatch.fnmatch(file_key, key):
                file_date = s3_hook.get_key(file_key, bucket).meta.data
                file_date = file_date["LastModified"]
                if int(interval) - abs((file_date - last_key[1]).days) >= 0:
                    last_key = (file_key, file_date)
        data = s3_hook.get_key(last_key[0], bucket)
        data.download_fileobj(result["data"])
        result["filename"] = last_key[0]
    except Exception as e:
        logger.error(f"Error trayendo datos desde S3 para el key {key}. {e}")
    return result


def get_base_date(conn: str, bucket: str, key: str) -> datetime.date:
    utc = pytz.UTC
    # Colocar una fecha muy atrás como base
    last_date = datetime.datetime(2000, 1, 1, 0, 0, 0).replace(tzinfo=utc)
    try:
        if key.rfind("/") != -1:
            prefix = key[:key.rfind("/") + 1]
        else:
            prefix = ""
        s3_hook = S3Hook(conn)
        files = s3_hook.list_keys(bucket, prefix)
        for file_key in files:
            if fnmatch.fnmatch(file_key, key):
                file_date = s3_hook.get_key(file_key, bucket).meta.data
                file_date = file_date["LastModified"]
                if last_date >= file_date:
                    continue
                last_date = file_date
        logger.debug(f"Fecha base desde {key} : {last_date}")
    except Exception as e:
        logger.error(f"Error buscando archivo base para tener la fecha base. key: {key}. {e}")
    return last_date


def save_df_to_s3(df: pd.DataFrame, conn: str, bucket: str, key: str, delimiter: str = ","):
    try:
        logger.info(f"SUBIENDO A NUBE KEY {key}")
        file_type = get_type_file(key)
        s3_hook = S3Hook(conn)
        if file_type == FileTypeEnum.EXCEL or file_type == FileTypeEnum.OLD_EXCEL:
            with BytesIO() as buffer:
                with pd.ExcelWriter(buffer, engine='xlsxwriter') as writer:
                    df.to_excel(writer, index=None)
            s3_hook.load_bytes(buffer.getvalue(), key, bucket, True)
        elif file_type == FileTypeEnum.CSV or file_type == FileTypeEnum.TEXT:
            csv_buffer = BytesIO()
            df.to_csv(csv_buffer, header=True, index=False, sep=delimiter, na_rep='None')
            csv_buffer.seek(0)
            s3_hook.load_bytes(csv_buffer.getvalue(), key, bucket, True)
    except Exception as e:
        logger.error(f"Error guardando archivos a S3. key: {key}. {e}")


def move_object_s3(conn: str, bucket: str, source_key: str, output_key: str):
    try:
        filename = source_key[source_key.rfind("/")+1:]
        output_key += filename
        s3_hook = S3Hook(conn)
        s3_hook.copy_object(source_key, output_key, bucket, bucket)
        s3_hook.delete_objects(bucket, source_key)
    except Exception as e:
        logger.error(f"Error moviendo archivo desde {source_key} hacia {output_key} en bucket {bucket}. {e}")
